/* File metServer_modMain.c */
/*
 ##############################################################
 # (C) 2015 BAE SYSTEMS
 #
 # ECOA Data Servers Sample - Weather Server
 #
 # $Id: metServer_modMain.c,v 2.3 2017/11/06 15:33:32 ksappleb Exp $
 #
 # Part of the BAE Systems, Rochester, ECOA Samples collection
 #
 # Created by K S Appleby, BAE Systems, Airport Works, Rochester, Kent
 #
 # This software is developed for and on behalf of BAE Systems (Operations)
 # Limited, Dassault Aviation, Bull SAS, Thales Systemes Aeroportes,
 # GE Aviation Systems Limited,
 # General Dynamics United Kingdom Limited and Leonardo MW Ltd,
 # and the copyright is owned by BAE Systems (Operations) Limited,
 # Dassault Aviation, Bull SAS, Thales Systemes Aeroportes, 
 # GE Aviation Systems Limited, General Dynamics United Kingdom
 # Limited and Leonardo MW Ltd
 #
 # This software is developed by BAE Systems (Operations)
 # Limited, Electronic Systems, and is the Intellectual
 # Property of BAE Systems (Operations) Limited, Electronic Systems.
 #
 # The information set out in this document is provided solely
 # on an 'as is' basis and the co-developers of this software make
 # no warranties expressed or implied, including no warranties
 # as to completeness, accuracy or fitness for purpose, with respect
 # to any of the information.
 ##############################################################
*/
#include "metServer_modMain.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <curl/curl.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>

/*
 *********************************************************************************************************************
 * Local types and forward references
 **********************************************************************************************************************
 */
typedef struct {
	int respCode;
	char respText[128];
} respHeader;

typedef struct {
	int size;
	char *data;
} respContent;

static int getWeather( metServer_modMain__context* context, const Weather__wxLocationString *location, Weather__WeatherRecd *weather );
static int getForecastWeather( metServer_modMain__context* context, ECOA__global_time when, const Weather__wxLocationString *location, Weather__WeatherRecd *fcast );
static int parseWeatherReport( metServer_modMain__context* context, const char* report, Weather__WeatherRecd *weather );
static int parseForecastReport( metServer_modMain__context* context, const char* report, const char *fcastDate, Weather__WeatherRecd *fcast );
static size_t response_handler( void *buffer, size_t size, size_t nmemb, void *userp );
static size_t header_handler( void *buffer, size_t size, size_t nmemb, void *userp );

static Weather__WeatherRecd wxNIL = (Weather__WeatherRecd){{0,0},"",0.0,0.0,0.0,"-",0.0,0.0,0.0,0.0,"Not available"};
static char *CONNECTIONFAILED = "Weather server connection error";

/*
 *********************************************************************************************************************
 * ECOA module operations
 **********************************************************************************************************************
 */
void metServer_modMain__INITIALIZE__received(metServer_modMain__context *context)
{
	context->user.wx = context->user.fcast = wxNIL;
}

void metServer_modMain__START__received(metServer_modMain__context *context)
{
   /* User Code Here */
}

void metServer_modMain__STOP__received(metServer_modMain__context *context)
{
   /* User Code Here */
}

void metServer_modMain__SHUTDOWN__received(metServer_modMain__context *context)
{
   /* User Code Here */
}

/**********************************************************************************************************************/

void metServer_modMain__setLocation__received(metServer_modMain__context *context,
    const Weather__wxLocationString *location)
{
	ECOA__global_time now;
	ECOA__log msg;
	ECOA__return_status errc;
	metServer_modMain_container__weather_handle wxHandle;
	/**/
	metServer_modMain_container__get_UTC_time( context, &now );
	//
	msg.current_size = sprintf( msg.data, "metServer_modMain__setLocation__received" );
//	metServer_modMain_container__log_trace( context, msg );
	/*
	 * If the location is not the same as the last access, OR the last access is obsolete,
	 * get a new set of data.
	 */
	if(( strcmp( (char*)context->user.wx.Location, (char*)location ) != 0 ) ||
		( now.seconds - context->user.wx.Timestamp.seconds >= 300 )){
			if( getWeather( context, (Weather__wxLocationString*)location, &(context->user.wx) ) == 0 ){
				context->user.wx.Timestamp = now;
				strcpy( context->user.wx.Location, *location );
				//
				errc = metServer_modMain_container__weather__get_write_access( context, &wxHandle );
				if( errc != ECOA__return_status_OK && errc != ECOA__return_status_DATA_NOT_INITIALIZED ){
					msg.current_size = sprintf( msg.data, "metServer_modMain_container__weather__get_write_access returned %d", errc );
					metServer_modMain_container__log_trace( context, msg );
				}else{
					(*wxHandle.data) = context->user.wx;
//					printf( "*** publishing weather with (e.g.) temp = %g\n", wxHandle.data->TemperatureC );
					metServer_modMain_container__weather__publish_write_access( context, &wxHandle );
				}
				msg.current_size = sprintf( msg.data, "metServer: setLocation: weather service accessed..." );
				metServer_modMain_container__log_trace( context, msg );
			}else{
				msg.current_size = sprintf( msg.data, CONNECTIONFAILED );
				metServer_modMain_container__log_warning( context, msg );
			}
	}
}

/**********************************************************************************************************************/

void metServer_modMain__getWeather__request_received
   (metServer_modMain__context* context,
    const ECOA__uint32 ID,
    const Weather__wxLocationString* location)
{
//	ECOA__global_time now;
	ECOA__log msg;
	Weather__WeatherRecd wx = wxNIL;
	//
	metServer_modMain_container__get_UTC_time( context, &(wx.Timestamp));
	strcpy( wx.Location, *location );
	if( getWeather( context, location, &wx ) == 0 ){
		msg.current_size = sprintf( msg.data, "metServer: getWeather: weather service accessed..." );
		metServer_modMain_container__log_trace( context, msg );
	}else{
		msg.current_size = sprintf( msg.data, CONNECTIONFAILED );
		metServer_modMain_container__log_warning( context, msg );
		sprintf( wx.Synopsis, "Not available" );
	}
	metServer_modMain_container__getWeather__response_send( context, ID, &wx );
}

/**********************************************************************************************************************/

void metServer_modMain__getForecastWeather__request_received
   (metServer_modMain__context* context,
    const ECOA__uint32 ID,
    const ECOA__global_time* when,
    const Weather__wxLocationString* location)
{
	ECOA__global_time now;
	ECOA__log msg;
	//
	msg.current_size = sprintf( msg.data, "metServer_modMain__getForecastWeather__request_received" );
	metServer_modMain_container__log_trace( context, msg );
	/**/
	if( getForecastWeather( context, *when, (Weather__wxLocationString*)location, &(context->user.fcast) ) != 0 ){
		msg.current_size = sprintf( msg.data, CONNECTIONFAILED );
		metServer_modMain_container__log_warning( context, msg );
	}else{
		metServer_modMain_container__get_UTC_time( context, &now );
		context->user.fcast.Timestamp = now;
		strcpy( context->user.fcast.Location, *location );
	}
	metServer_modMain_container__getForecastWeather__response_send( context, ID, &(context->user.fcast) );
}

/**********************************************************************************************************************/

void metServer_modMain__getWeatherItem__request_received
   (metServer_modMain__context* context,
    const ECOA__uint32 ID,
    const Weather__wxWeatherItem item,
    const Weather__wxLocationString* location)
{
	ECOA__global_time now;
	ECOA__log msg;
	Weather__wxItemUnion wxItem;
	//
	msg.current_size = sprintf( msg.data, "metServer_modMain__getWeatherItem__request_received" );
	metServer_modMain_container__log_trace( context, msg );
	/*
	 * For the individual data items, only refresh if the lifetime has expired...
	 */
	metServer_modMain_container__get_UTC_time( context, &now );
	if( context->user.wx.Timestamp.seconds + WX_LIFETIME < now.seconds ){
		if( location == NULL || strlen( (char*)location ) == 0 ){
			if( getWeather( context, &(context->user.wx.Location), &(context->user.wx) ) == 0 ){
				context->user.wx.Timestamp = now;
				strcpy( context->user.wx.Location, *location );
			}else{
				msg.current_size = sprintf( msg.data, CONNECTIONFAILED );
				metServer_modMain_container__log_warning( context, msg );
			}
		}else{
			if( getWeather( context, (Weather__wxLocationString*)location, &(context->user.wx) ) == 0 ){
				context->user.wx.Timestamp = now;
				strcpy( context->user.wx.Location, *location );
			}else{
				msg.current_size = sprintf( msg.data, CONNECTIONFAILED );
				metServer_modMain_container__log_warning( context, msg );
			}
		}
	}
	wxItem.param = item;
	switch( item ){
		case Weather__wxWeatherItem_TEMP:
			wxItem.u_param.TemperatureC = context->user.wx.TemperatureC; break;
		case Weather__wxWeatherItem_MINTEMP:
			wxItem.u_param.MinTempC = context->user.wx.MinTempC; break;
		case Weather__wxWeatherItem_MAXTEMP:
			wxItem.u_param.MaxtempC = context->user.wx.MaxtempC; break;
		case Weather__wxWeatherItem_WINDIR:
			strcpy( wxItem.u_param.WindDir, context->user.wx.WindDir ); break;
		case Weather__wxWeatherItem_WINSPD:
			wxItem.u_param.WindSpeedMPH = context->user.wx.WindSpeedMPH; break;
		case Weather__wxWeatherItem_PRESS:
			wxItem.u_param.PressureMB = context->user.wx.PressureMB; break;
		case Weather__wxWeatherItem_CLOUD:
			wxItem.u_param.CloudCoverPCNT = context->user.wx.CloudCoverPCNT; break;
		case Weather__wxWeatherItem_PRECIP:
			wxItem.u_param.PreciptationMM = context->user.wx.PreciptationMM; break;
		case Weather__wxWeatherItem_SYNOP:
			strcpy( wxItem.u_param.Synopsis, context->user.wx.Synopsis ); break;
	}
	metServer_modMain_container__getWeatherItem__response_send( context, ID, &wxItem );
}

/**********************************************************************************************************************/

void metServer_modMain__getForecastWeatherItem__request_received
   (metServer_modMain__context* context,
    const ECOA__uint32 ID,
    const Weather__wxWeatherItem item,
    const ECOA__global_time* when,
    const Weather__wxLocationString* location)
{
	ECOA__global_time now;
	ECOA__log msg;
	Weather__wxItemUnion wxItem;
	//
	msg.current_size = sprintf( msg.data, "metServer_modMain__getForecastWeatherItem__request_received" );
	metServer_modMain_container__log_trace( context, msg );
	/*
	 * For the individual data items, only refresh if the lifetime has expired...
	 */
	metServer_modMain_container__get_UTC_time( context, &now );
	if( context->user.fcast.Timestamp.seconds + WX_LIFETIME < now.seconds ){
		if( getForecastWeather( context, *when, (Weather__wxLocationString*)location, &(context->user.fcast) ) != 0 ){
			msg.current_size = sprintf( msg.data, CONNECTIONFAILED );
			metServer_modMain_container__log_warning( context, msg );
		}else{
			context->user.fcast.Timestamp = now;
			strcpy( context->user.fcast.Location, *location );
		}
	}
	wxItem.param = item;
	switch( item ){
		case Weather__wxWeatherItem_TEMP:
			wxItem.u_param.TemperatureC = context->user.fcast.TemperatureC; break;
		case Weather__wxWeatherItem_MINTEMP:
			wxItem.u_param.MinTempC = context->user.fcast.MinTempC; break;
		case Weather__wxWeatherItem_MAXTEMP:
			wxItem.u_param.MaxtempC = context->user.fcast.MaxtempC; break;
		case Weather__wxWeatherItem_WINDIR:
			strcpy( wxItem.u_param.WindDir, context->user.fcast.WindDir ); break;
		case Weather__wxWeatherItem_WINSPD:
			wxItem.u_param.WindSpeedMPH = context->user.fcast.WindSpeedMPH; break;
		case Weather__wxWeatherItem_PRESS:
			wxItem.u_param.PressureMB = context->user.fcast.PressureMB; break;
		case Weather__wxWeatherItem_CLOUD:
			wxItem.u_param.CloudCoverPCNT = context->user.fcast.CloudCoverPCNT; break;
		case Weather__wxWeatherItem_PRECIP:
			wxItem.u_param.PreciptationMM = context->user.fcast.PreciptationMM; break;
		case Weather__wxWeatherItem_SYNOP:
			strcpy( wxItem.u_param.Synopsis, context->user.fcast.Synopsis ); break;
	}
	metServer_modMain_container__getForecastWeatherItem__response_send( context, ID, &wxItem );
}

/**********************************************************************************************************************/

void metServer_modMain__Tick__received(metServer_modMain__context *context)
{
	ECOA__return_status errc;
	ECOA__global_time now;
	ECOA__log msg;
	metServer_modMain_container__weather_handle wxHandle;
	Weather__wxLocationString location;
	/**/
	metServer_modMain_container__get_UTC_time( context, &now );
	if(( context->user.wx.Location != NULL && strlen( context->user.wx.Location ) != 0 ) ){
		strcpy( location, context->user.wx.Location );
		/*
		 * If the last access is obsolete, get a new set of data.
		 */
		if( now.seconds - context->user.wx.Timestamp.seconds >= 300 ){
			if( getWeather( context, &location, &(context->user.wx) ) == 0 ){
				metServer_modMain_container__get_UTC_time( context, &now );
				context->user.wx.Timestamp = now;
				strcpy( context->user.wx.Location, location );
				/**/
				msg.current_size = sprintf( msg.data, "metServer: periodic update: weather service accessed..." );
				metServer_modMain_container__log_trace( context, msg );
			}else{
				msg.current_size = sprintf( msg.data, CONNECTIONFAILED );
				metServer_modMain_container__log_warning( context, msg );
			}
		}else{
			msg.current_size = sprintf( msg.data, "metServer: weather service NOT accessed this time around..." );
//			metServer_modMain_container__log_trace( context, msg );
		}
	}else{
		context->user.wx.Timestamp = now;
	}
	errc = metServer_modMain_container__weather__get_write_access( context, &wxHandle );
	if( errc != ECOA__return_status_OK && errc != ECOA__return_status_DATA_NOT_INITIALIZED ){
		msg.current_size = sprintf( msg.data, "metServer_modMain_container__weather__get_write_access returned %d", errc );
		metServer_modMain_container__log_trace( context, msg );
	}else{
		(*wxHandle.data) = context->user.wx;
//		printf( "*** publishing weather with (e.g.) synopsis = %s\n", wxHandle.data->Synopsis );
		metServer_modMain_container__weather__publish_write_access( context, &wxHandle );
	}
}

/*
 *********************************************************************************************************************
 * Local operations...
 *********************************************************************************************************************
 */

static int getWeather( metServer_modMain__context* context, const Weather__wxLocationString *location, Weather__WeatherRecd *weather )
{
	char request[128];
	CURLcode    res;
	CURL       *easyHandle;
	respHeader  respHdr;
	respContent respData = {0, NULL};
	int         retv = 0;
	ECOA__log   msg;
	//
	curl_global_init( CURL_GLOBAL_ALL );
	//
	easyHandle = curl_easy_init();
	//
	// Handle response data
	curl_easy_setopt( easyHandle, CURLOPT_HEADER, 0L ); // Make sure header does NOT get treated as "response"
	curl_easy_setopt( easyHandle, CURLOPT_HEADERFUNCTION, header_handler );
	curl_easy_setopt( easyHandle, CURLOPT_HEADERDATA, &respHdr );
	curl_easy_setopt( easyHandle, CURLOPT_WRITEFUNCTION, response_handler );
	curl_easy_setopt( easyHandle, CURLOPT_WRITEDATA, &respData );
	//
	// Some servers don't like requests that are made without a user-agent field, so we provide one
	curl_easy_setopt( easyHandle, CURLOPT_USERAGENT, "libcurl-agent/1.0" );
	//
	//#############################################################
	// Control the proxy config using environment variables...
	//   set http_proxy=http://username:password@proxyip:proxyport
	//   set no_proxy=host{,host|.domain...}
	//#############################################################
	//
	// What do we want to read?
	sprintf( request, "http://www.myweather2.com/developer/forecast.ashx?uac=%s&query=%s", MYWEATHER2_UAC, *location );
	curl_easy_setopt( easyHandle, CURLOPT_URL, request );
	//
	// Do it...
	res = curl_easy_perform( easyHandle );
	if( res != CURLE_OK ) {
	    msg.current_size = sprintf( msg.data, "metServer: curl_easy_perform() failed: %s", curl_easy_strerror(res) );
		metServer_modMain_container__log_warning( context, msg );
		retv = -1;
	}
	if( respHdr.respCode != 200 ){
		msg.current_size = sprintf( msg.data, "metServer: Weather service returned \"%d %s\"", respHdr.respCode, respHdr.respText );
		metServer_modMain_container__log_warning( context, msg );
		retv = -1;
	}else{
//		printf( "*** received %d bytes from %s\n%s\n", respData.size, request, respData.data );
		if( parseWeatherReport( context, respData.data, weather ) != 0 ){
			msg.current_size = sprintf( msg.data, "metServer: Invalid weather report from service" );
			metServer_modMain_container__log_warning( context, msg );
			retv = 0;
		}
//		printf( "*** received weather report with temp = %g\n", weather->TemperatureC );
	}
	//
	curl_easy_cleanup( easyHandle );
	return retv;
}

/**********************************************************************************************************************/

static int getForecastWeather( metServer_modMain__context* context, ECOA__global_time when, const Weather__wxLocationString *location, Weather__WeatherRecd *fcast )
{
	char request[128];
	CURLcode    res;
	CURL       *easyHandle;
	respHeader  respHdr;
	respContent respData = {0, NULL};
	int         retv = 0;
	ECOA__global_time now;
	ECOA__log   msg;
	struct tm   tmWhen;
	time_t      when_seconds = when.seconds;
	//
	memcpy( &tmWhen, localtime(&(when_seconds)), sizeof(struct tm) );
	//
	// The weather service I'm using only gives forecasts for tomorrow and the day after - unless I pay cash money!
	// So if the request is for a longer term than that, apologise and bomb out.
	if(( when.seconds < now.seconds ) || ( when.seconds - now.seconds) > 172800 ){
		msg.current_size = sprintf( msg.data, "metServer: no forecast available for %s", asctime(&tmWhen));
		metServer_modMain_container__log_warning( context, msg );
		retv = -1;
	}else{
		curl_global_init( CURL_GLOBAL_ALL );
		//
		easyHandle = curl_easy_init();
		//
		// Handle response data
		curl_easy_setopt( easyHandle, CURLOPT_HEADER, 0L ); // Make sure header does NOT get treated as "response"
		curl_easy_setopt( easyHandle, CURLOPT_HEADERFUNCTION, header_handler );
		curl_easy_setopt( easyHandle, CURLOPT_HEADERDATA, &respHdr );
		curl_easy_setopt( easyHandle, CURLOPT_WRITEFUNCTION, response_handler );
		curl_easy_setopt( easyHandle, CURLOPT_WRITEDATA, &respData );
		//
		// Some servers don't like requests that are made without a user-agent field, so we provide one
		curl_easy_setopt( easyHandle, CURLOPT_USERAGENT, "libcurl-agent/1.0" );
		//
		//#############################################################
		// Control the proxy config using environment variables...
		//   set http_proxy=http://username:password@proxyip:proxyport
		//   set no_proxy=host{,host|.domain...}
		//#############################################################
		//
		// What do we want to read?
		sprintf( request, "http://www.myweather2.com/developer/forecast.ashx?uac=%s&query=%s", MYWEATHER2_UAC, *location );
		curl_easy_setopt( easyHandle, CURLOPT_URL, request );
		//
		// Do it...
		res = curl_easy_perform( easyHandle );
		if( res != CURLE_OK ) {
			msg.current_size = sprintf( msg.data, "metServer: curl_easy_perform() failed: %s", curl_easy_strerror(res) );
			metServer_modMain_container__log_warning( context, msg );
		}
		if( respHdr.respCode != 200 ){
			msg.current_size = sprintf( msg.data, "metServer: Weather service returned \"%d %s\"", respHdr.respCode, respHdr.respText );
			metServer_modMain_container__log_warning( context, msg );
		}else{
//			printf( "*** received %d bytes from %s\n%s\n", respData.size, request, respData.data );
			// Do we want the 24 or 48 hour forecast?
			char fcastDate[32];
			sprintf( fcastDate, "%4d-%02d-%02d", tmWhen.tm_year, tmWhen.tm_mon, tmWhen.tm_mday);
			if( parseForecastReport( context, respData.data, fcastDate, fcast ) != 0 ){
				msg.current_size = sprintf( msg.data, "metServer: Invalid weather report from service" );
				metServer_modMain_container__log_warning( context, msg );
			}
//			printf( "*** received weather report with temp = %g\n", weather->TemperatureC );
		}
		//
		curl_easy_cleanup( easyHandle );
	}
	return retv;
}

/**********************************************************************************************************************/

static int parseWeatherReport( metServer_modMain__context* context, const char* report, Weather__WeatherRecd *weather )
{
	xmlDocPtr  pDoc = NULL;
	xmlXPathContextPtr xpathCtx;
	xmlXPathObjectPtr xpathObj;
	xmlChar* p;
	ECOA__double64 windSpeed;
	//
	if(( pDoc = xmlParseMemory( report, strlen( report ))) == NULL ){
		return -1;
	}
	if(( xpathCtx = xmlXPathNewContext(pDoc)) == NULL ){
        xmlFreeDoc(pDoc);
		return -1;
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/temp", xpathCtx)) != NULL ){
		weather->TemperatureC = atof( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/wind/speed", xpathCtx)) != NULL ){
		windSpeed = atof( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/wind/wind_unit", xpathCtx)) != NULL ){
		if( strcasecmp( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )), "MPH" ) == 0 ){
			weather->WindSpeedMPH = windSpeed;
		}else
			// scale from km/h
			weather->WindSpeedMPH = windSpeed / 1.609;
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/wind/dir", xpathCtx)) != NULL ){
		p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] );
		if( xmlStrlen( p ) <= 3 )
			strcpy( weather->WindDir, (char*)p );
		else
			strcpy( weather->WindDir, "-" );
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/pressure", xpathCtx)) != NULL ){
		weather->PressureMB = atof( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/weather_text", xpathCtx)) != NULL ){
		strcpy( weather->Synopsis, (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	xmlXPathFreeContext( xpathCtx );
	xmlFreeDoc( pDoc );
	return 0;
}

/**********************************************************************************************************************/

static int parseForecastReport( metServer_modMain__context* context, const char* report, const char *fcastDate, Weather__WeatherRecd *fcast )
{
#if 0
	xmlDocPtr  pDoc = NULL;
	xmlXPathContextPtr xpathCtx;
	xmlXPathObjectPtr xpathObj;
	xmlChar* p;
	ECOA__double64 windSpeed;
	//
	if(( pDoc = xmlParseMemory( report, strlen( report ))) == NULL ){
		return -1;
	}
	if(( xpathCtx = xmlXPathNewContext(pDoc)) == NULL ){
        xmlFreeDoc(pDoc);
		return -1;
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/temp", xpathCtx)) != NULL ){
		weather->TemperatureC = atof( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/wind/speed", xpathCtx)) != NULL ){
		windSpeed = atof( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/wind/wind_unit", xpathCtx)) != NULL ){
		if( strcasecmp( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )), "MPH" ) == 0 ){
			weather->WindSpeedMPH = windSpeed;
		}else
			// scale from km/h
			weather->WindSpeedMPH = windSpeed / 1.609;
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/wind/dir", xpathCtx)) != NULL ){
		strcpy( weather->WindDir, (char*)( p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/pressure", xpathCtx)) != NULL ){
		weather->PressureMB = atof( (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	if(( xpathObj = xmlXPathEval((xmlChar*)"//curren_weather/weather_text", xpathCtx)) != NULL ){
		strcpy( weather->Synopsis, (char*)(p = xmlNodeGetContent( xpathObj->nodesetval->nodeTab[0] )));
		xmlFree(p);
		xmlXPathFreeObject( xpathObj );
	}
	xmlXPathFreeContext( xpathCtx );
	xmlFreeDoc( pDoc );
#else
	/* TODO */
	fcast->MaxtempC = fcast->MinTempC = fcast->PressureMB = fcast->TemperatureC = fcast->WindSpeedMPH = 0.0;
	fcast->Synopsis[0] = fcast->WindDir[0] = '\000';
#endif
	return 0;
}

/*
 *********************************************************************************************************************
 * Local operations for libcurl functions...
 *********************************************************************************************************************
 */

static size_t response_handler( void *buffer, size_t size, size_t nmemb, void *userp )
{
	respContent *response = userp;
	/**/
	if( response->data == NULL ){
		response->data = (char*)malloc( size*nmemb + 1 );
		memcpy( response->data, buffer, (response->size = size*nmemb));
		response->data[response->size] = '\000';
	}else{
		char *p;
		int newSize = response->size + size*nmemb;
		response->data = (char*)realloc( response->data, newSize );
		p = (response->data) + response->size;
		memcpy( p, buffer, nmemb*size );
		response->size = newSize;
		response->data[newSize] = '\000';
	}
//	printf( "Response: %s\n", (char*)buffer );
//	printf( "Response: %s\n", response );
	//
	// A response handler needs to return the number of response bytes processed...
	return ( size * nmemb );
}

static size_t header_handler( void *buffer, size_t size, size_t nmemb, void *userp )
{
	int res, majVer, minVer;
	respHeader *respHdr = (respHeader*)userp;
	char tmp[128];
	//
//	printf( "Header: %s\n", (char*)buffer );
	res = sscanf( (char*)buffer, "HTTP/%d.%d %d", &majVer, &minVer, &(respHdr->respCode) );
	if( res == 3 ){
		sprintf( tmp, "HTTP/%d.%d %d ", majVer, minVer, respHdr->respCode );
		strcpy( respHdr->respText, strtok( ((char*)buffer + strlen( tmp )), "\r\n" ));
//		printf( "Header: Code = %d '%s'\n", *respCode, respText );
	}
	// A response/header handler needs to return the number of response bytes processed...
	return ( size * nmemb );
}
